package ap.proof.goal;

import ap.terfor.{Formula, Conjunction, ArithConj, Quantifier}
import ap.util.Debug

object ComplexFormulaTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK

  
}

case class ComplexFormulaTask(formula : Formula) extends Task {
  
  val priority : Int = -20000

  def apply(goal : Goal) : TaskResult = {
    if (formula.isTrue) {
      // then the goal can be closed immediately
      UpdateGoalTaskResult(Conjunction.FALSE, goal)
    } else if (formula.isFalse) {
      // then the formula can be removed
      UnchangedGoalTaskResult
    } else {
      // the complex formula has to be examined and possibly split into
      // multiple conjunctions (but without splitting the goal, this is done
      // by <code>ConjunctionTask</code>)
  
      val order = goal.order
      
      def extendedFacts(newFacts : Formula) : Conjunction =
        Conjunction.conj(Array(goal.facts, newFacts), order)
      
      def addLiteral(lit : Formula) : TaskResult =
        UpdateGoalTaskResult(extendedFacts(Conjunction.negate(lit, order)),
                             Array(FactsNormalisationTask),
                             goal)        

      def addTaskFor(conj : Conjunction) : TaskResult =
        UpdateGoalTaskResult(Array(createTaskFor(conj, goal.age)), goal)

      formula match {
        
      case ac : ArithConj =>
      	if (ac.isLiteral)
          addLiteral(ac)
      	else
          addTaskFor(Conjunction.conj(ac, order))

      case conj : Conjunction =>
        if (conj.isLiteral) {
          addLiteral(conj)
        } else if (conj.isNegatedConjunction) {
          // apply alpha-rules
          
          val disj = conj.negatedConjs(0)
          //////////////////////////////////////////////////////////////////////
          // possibly existing quantifiers should have been pulled out and
          // should not occur at this point
          Debug.assertInt(ComplexFormulaTask.AC, disj.quans.isEmpty)
          //////////////////////////////////////////////////////////////////////
          
          // the literals of the negated conjunction can be added as facts to
          // the goal
          val updatedFacts = extendedFacts(disj.arithConj)
          
          // the nested conjunctions have to be handled recursively
          val newTasks = for (f <- disj.negatedConjs) yield new ComplexFormulaTask(f)
          
	  UpdateGoalTaskResult(updatedFacts,
                               newTasks ++ Array(FactsNormalisationTask),
                               goal)
        } else {
          addTaskFor(conj)
        }
      }
    }
  }
  
  private def createTaskFor(conj : Conjunction, age : Int) =
    if (conj.quans.isEmpty) {
      new ConjunctionTask(conj, age)
    } else conj.quans.last match {
      case Quantifier.ALL => new AllQuantifierTask(conj, age)
      case Quantifier.EX => new ExQuantifierTask(conj, age)
    }
}
