package ap.terfor.equations;

import ap.terfor.linearcombination.LinearCombination
import ap.basetypes.IdealInt
import ap.util.{Debug, Logic, Seqs}

object EquationSet {

  private val AC = Debug.AC_EQUATIONS
  
  
}

abstract class EquationSet protected (private val lhss : Array[LinearCombination],
                                      val order : TermOrder)
               extends Formula with RandomAccessSeq[LinearCombination] {
  
  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(EquationSet.AC,
                   (
                     // as a special case, we allow a singleton set with a
                     // constant as element (to be able to express unsatisfiable
                     // conjunctions or valid disjunctions of equations properly)
                     size == 1 && (this(0) == LinearCombination.ZERO ||
                                   this(0) == LinearCombination.ONE) 
                     ||
                     // otherwise, only primitive elements are allowed
                     Logic.forall(for (lhs <- this.elements)
                                  yield (lhs.isPrimitive && lhs.isPositive))
                   )
                   &&
                   Logic.forall(0, this.size - 1,
                                (i:Int) => order.compare(this(i), this(i+1)) > 0))
  //////////////////////////////////////////////////////////////////////////////

  def length : Int = lhss.length
    
  def apply(i : Int) : LinearCombination = lhss(i)
    
  override def elements = lhss.elements

  //////////////////////////////////////////////////////////////////////////////

  lazy val variables : Set[VariableTerm] =
    Set.empty ++ (for (lc <- this.elements; v <- lc.variables.elements) yield v)

  lazy val constants : Set[ConstantTerm] =
    Set.empty ++ (for (lc <- this.elements; c <- lc.constants.elements) yield c)
      
  //////////////////////////////////////////////////////////////////////////////

  protected def relationString : String
  
  override def toString : String = {
    if (isTrue) {
      "true"
    } else if (isFalse) {
      "false"
    } else {
      val strings = for (lhs <- this.elements)
                    yield ("" + lhs + " " + relationString + " 0")
      if (strings.hasNext)
        strings.reduceLeft((s1 : String, s2 : String) =>
                           s1 + " & " + s2)
      else
        throw new Error // should never be reached
    }
  }
  
  override def equals(that : Any) : Boolean = that match {
    case that : EquationSet => lhss.deepEquals(that.lhss)
    case _ => false
  }

  private lazy val hashCodeVal = Seqs.computeHashCode(this, 0, 3)

  override def hashCode = hashCodeVal
}
